//===- IRBindings.h - Additional bindings for IR ----------------*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file defines additional C bindings for the IR component.
//
//===----------------------------------------------------------------------===//

/*
 * Shamelessly based on IRBindings.h from LLVM's GO bindings. The original
 * file could be found at bindings/go/llvm/IRBindings.h in LLVM source tree.
 */

#ifndef LLVM_BINDINGS_GO_LLVM_IRBINDINGS_H
#define LLVM_BINDINGS_GO_LLVM_IRBINDINGS_H

struct LLVMOpaqueMetadata;

#if defined(__cplusplus) && !defined(__GCCXML__)
#include "llvm-c/Core.h"
#include "llvm-c/ExecutionEngine.h"
#include "llvm/IR/Metadata.h"
#include "llvm/Support/CBindingWrapping.h"
#endif

#include <stdint.h>

#include <llvm-c/Core.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef struct LLVMOpaqueMetadata *LLVMMetadataRef;

// These functions duplicate the LLVM*FunctionAttr functions in the stable C
// API. We cannot use the existing functions because they take 32-bit attribute
// values, and the Go bindings expose all of the LLVM attributes, some of which
// have values >= 1<<32.

void LLVMAddFunctionAttr2(LLVMValueRef Fn, uint64_t PA);
uint64_t LLVMGetFunctionAttr2(LLVMValueRef Fn);
void LLVMRemoveFunctionAttr2(LLVMValueRef Fn, uint64_t PA);

LLVMMetadataRef LLVMConstantAsMetadata(LLVMValueRef Val);

LLVMMetadataRef LLVMMDString2(LLVMContextRef C, const char *Str, unsigned SLen);
LLVMMetadataRef LLVMMDNode2(LLVMContextRef C, LLVMMetadataRef *MDs,
                            unsigned Count);
LLVMMetadataRef LLVMTemporaryMDNode(LLVMContextRef C, LLVMMetadataRef *MDs,
                                    unsigned Count);

void LLVMAddNamedMetadataOperand2(LLVMModuleRef M, const char *name,
                                  LLVMMetadataRef Val);

/**
 * Set the metadata of the specified kind to the specified
 * node. This updates/replaces metadata if already present, or 
 * removes it if `MD` is null. 
 * 
 * `Val` mist be either an instruction or a function. If it;s of 
 *  other kind. this function is no-op
 *
 *  \param Val reference either an instruction or a function.
 *  \param KindID a unique identifier of metadata kind. 
 *         \see llvm::LLVMContext
 *
 *  \see llvm::LLVMContext
 *  \see llvm::Instruction::setMetadata()
 *  \see llvm::Function::setMetadata()
 */

void LLVMSetMetadata2(LLVMValueRef Val, unsigned KindID, LLVMMetadataRef MD);

void LLVMMetadataReplaceAllUsesWith(LLVMMetadataRef MD, LLVMMetadataRef New);

/**
 * Return a user-friendly representation of the metadata node.
 * Usefull mainly for debugging. 
 *
 * \return string representation of the metadata. The caller is
 *         responsible for free()ing the string
 */
char* LLVMPrintMetadataToString(LLVMMetadataRef MD);


/**
 * Set the current source location in the original source. This information
 * is used to attach debug information to each instruction generated
 * from now onwards (until a new location is set)
 *
 * \param Bref the IR builder in which to set the location
 * \param Line source line of upcoming instructions
 * \param Col source column of upcoming instructions
 * \param Scope a scope in which the location belongs, i.e.,
 *   a DISubroutine, a DILexicalBlock end so on.
 * \param InlinedAt if the code is inlined, the scope in which
 *   it is inlined
 */
void LLVMSetCurrentDebugLocation2(LLVMBuilderRef Bref, unsigned Line,
                                  unsigned Col, LLVMMetadataRef Scope,
                                  LLVMMetadataRef InlinedAt);

/**
 * Returns current source location as LLVM metadata (instance of DILocation).
 * Reference to the location may be then passed to LLVMDiBuilderInsertDeclare()
 * and LLVMDiBuilderInsertValue().
 *
 * \param Bref a builder from which to obtain the location
 * \return DILocation as LLVMMetadataRef
 *
 * \see LLVMDiBuilderInsertDeclare()
 * \see LLVMDiBuilderInsertValue()
 */
LLVMMetadataRef LLVMGetCurrentDebugLocation2(LLVMBuilderRef Bref);


/**
 * This enumeration defines the supported behaviors of module flags.
 *
 * \see llvm::Module::ModFlagBehavior
 * \see llbm::Module::addModuleFlag()
 * \see LLVMAddModuleFlag()
 */
enum LLVMModuleFlagBehavior {
	/**
	 * Emits an error if two values disagree, otherwise the resulting value is
	 * that of the operands.
	 */
	LLVMModuleFlagBehaviorError = 1,

	/**
	 * Emits a warning if two values disagree. The result value will be the
	 * operand for the flag from the first module being linked.
	 */
	LLVMModuleFlagBehaviorWarning = 2,

	/**
	 * Adds a requirement that another module flag be present and have a
	 * specified value after linking is performed. The value must be a metadata
	 * pair, where the first element of the pair is the ID of the module flag
	 * to be restricted, and the second element of the pair is the value the
	 * module flag should be restricted to. This behavior can be used to
	 * restrict the allowable results (via triggering of an error) of linking
	 * IDs with the **Override** behavior.
	 */
	LLVMModuleFlagBehaviorRequire = 3,

	/**
	 * Uses the specified value, regardless of the behavior or value of the
	 * other module. If both modules specify **Override**, but the values
	 * differ, an error will be emitted.
	 */
	LLVMModuleFlagBehaviorOverride = 4,

	/**
	 * Appends the two values, which are required to be metadata nodes.
	 */
	LLVMModuleFlagBehaviorAppend = 5,

	/**
	 * Appends the two values, which are required to be metadata
	 * nodes. However, duplicate entries in the second list are dropped
	 * during the append operation.
	 */
	LLVMModuleFlagBehaviorAppendUnique = 6,
};

/**
 * Adds a flag to the module.
 *
 * \see llvm::Module::addModuleFlag()
 */
void LLVMModuleAddModuleFlag(LLVMModuleRef M, LLVMModuleFlagBehavior Behavior, const char *Key, LLVMMetadataRef Val);

/**
 * @defgroup metadatatypec Metadata - Types
 *
 * Set of functions to query metadata node type (such as
 * DILocation, DIFile and so on
 */

/**
 * Metadata kind ID definition
 *
 * \see llvm::MetadataKind
 * \see LLVMGetMetadataKind()
 */

enum LLVMMetadataKind {
    /* 
     * If adding new entries, ALWAYS add them at the end,
     * otherwise a binary compatibility will be broken! 
     */
    MDTupleKind = 0,
    DILocationKind = 1,
    GenericDINodeKind = 2,
    DISubrangeKind = 3,
    DIEnumeratorKind = 4,
    DIBasicTypeKind = 5,
    DIDerivedTypeKind = 6,
    DICompositeTypeKind = 7,
    DISubroutineTypeKind = 8,
    DIFileKind = 9,
    DICompileUnitKind = 10,
    DISubprogramKind = 11,
    DILexicalBlockKind = 12,
    DILexicalBlockFileKind = 13,
    DINamespaceKind = 14,
    DIModuleKind = 15,
    DITemplateTypeParameterKind = 16,
    DITemplateValueParameterKind = 17,
    DIGlobalVariableKind = 18,
    DILocalVariableKind = 19,
    DIExpressionKind = 20,
    DIObjCPropertyKind = 21,
    DIImportedEntityKind = 20,
    ConstantAsMetadataKind = 23,
    LocalAsMetadataKind = 24,
    MDStringKind = 25,
    DIMacroKind = 26,
    DIMacroFileKind = 27,

    DistinctMDOperandPlaceholderKind = 28
    /* 
     * Add new entries only at the end!
     */
};

/**
 * By default, only sections that are "required for execution" are passed to the
 * memory manager, and other sections are discarded.
 * 
 * Passing 'true' to this method will cause executiuon engine to pass all 
 * sections to its memory manager regardless of whether they are "required to 
 * execute" in the usual sense.
 *
 * \see llvm::ExecutionEngine::setProcessAllSections(cool)
 */
void LLVMSetProcessAllSections(LLVMExecutionEngineRef EE, bool ProcessAllSections);

/**
 * Return metadata kind ID for given metadata node
 *
 * \param MD metadata node whose kind ID is queried
 * \return Metadata kind ID
 *
 * \see llvm::MetadataKind
 * \see LLVMMetadataKind
 *
 */
LLVMMetadataKind LLVMGetMetadataKind(LLVMMetadataRef MD);


/**
 * }
 */ // end of @defgroup metadatatypec Metadata - Types



/**
 * @defgroup metadataconv Metadata - Value Conversions
 *
 * Set of functions to convert metadata to values and back.
 * These are useful in cases when to "call" an intrinsic function with
 * metadata as parameters, such as @llvm.dbg.declare.
 */

/**
 * Convert metadata node to it's value representation.
 *
 * \param MD metadata to convert
 * \return metadata represented as LLVM value
 *
 * \see LLVMMetadataAsValueInContext()
 */
LLVMValueRef LLVMMetadataAsValue(LLVMMetadataRef MD);

/**
 * Convert metadata node to it's value representation in given context.
 *
 * \param C context in which the conversion takes place
 * \param MD metadata to convert
 * \return metadata represented as LLVM value
 *
 * \see LLVMMetadataAsValue()
 */
LLVMValueRef LLVMMetadataAsValueInContext(LLVMContextRef C, LLVMMetadataRef MD);


/**
 * Converts value to metadata.
 *
 * \param Value value to convert
 * \return metadata representation of the talue
 */
LLVMMetadataRef LLVMValueAsMetadata(LLVMValueRef Value);

/**
 * }
 */ // end of @defgroup metadataconv Metadata - Value Conversions


/**
 * @defgroup constvalconv Constant value conversions
 *
 * A set of functions to convert constant values represented
 * as LLVMValueRef to numeric values.
 *
 * If the LLVMValueRef is not a constant value of requested type,
 * 0 is returned. Since 0 is also a valid value, caller should
 * first check whether the value is actually a constant value using
 * LLVMIsConstant() and check whether the type matches:
 *
 * @code{.c}
 * ...
 * LLVMValueRef valueRef = ...;
 * int32_t value;
 * if( LLVMIsConstant(valueRef) &&
 *     LLVMGetTypeKind(LLVMTypeOf(valueRef)) == LLVMIntegerTypeKind &&
 *     LLVMGetIntTypeWidth(LLVMTypeOf(valueRef)) <= 64) {
 *     value = LLVMValueAsSInt64();
 * } else {
 *     ...valurRef is not a constant value of type int64_t...
 * }
 *
 * @endcode
 *
 * @{
 */

/**
 * Returns a signed 64-bit integer value of given LLVMValueRef.
 *
 * \param Value a (constant) value of type i64 (or narrower)
 * \return signed 64-bit integer value that passed Value encapsulates
 *
 * \see constvalconv
 */
int64_t LLVMValueAsSInt64(LLVMValueRef Value);

/**
 * Returns a unsigned 64-bit integer value of given LLVMValueRef.
 *
 * \param Value a (constant) value of type i64 (or narrower)
 * \return unsigned 64-bit integer value that passed Value encapsulates
 *
 * \see constvalconv
 */
uint64_t LLVMValueAsUInt64(LLVMValueRef Value);

/* TODO: Add functions for FP types and perhaps for strings.   */

/**
 * }
 */ // end of @defgroup constvalconv Constant value conversions


/**
 * @defgroup irbuilderintrin Intrinsics
 *
 * Set of builder functions to make use of various 
 * intrinsics. 
 */
 
/** 
 * Create and insert a memset to the specified pointer and the
 * specified value.
 *
 * If the pointer isn't an i8*, it will be converted.
 */
LLVMValueRef LLVMBuildMemSet(LLVMBuilderRef B, LLVMValueRef Ptr, LLVMValueRef Val, LLVMValueRef Size, uint32_t Align, uint32_t IsVolatile);


/**
  * Create and insert a memcpy between the specified pointers.
  *
  * If the pointers aren't i8*, they will be converted. 
  */
LLVMValueRef LLVMBuildMemCpy(LLVMBuilderRef B, LLVMValueRef Dst, LLVMValueRef Src, LLVMValueRef Size, uint32_t Align, uint32_t IsVolatile);  

/**
  * Create and insert a memmove between the specified pointers.
  *
  * If the pointers aren't i8*, they will be converted. 
  */
LLVMValueRef LLVMBuildMemMove(LLVMBuilderRef B, LLVMValueRef Dst, LLVMValueRef Src, LLVMValueRef Size, uint32_t Align, uint32_t IsVolatile);  


/**
 * }
 */ // end of @defgroup irbuilderintrin Intrinsics


#ifdef __cplusplus
}
#endif

#if defined(__cplusplus) && !defined(__GCCXML__)

namespace llvm {

DEFINE_ISA_CONVERSION_FUNCTIONS(Metadata, LLVMMetadataRef)

inline Metadata **unwrap(LLVMMetadataRef *Vals) {
  return reinterpret_cast<Metadata**>(Vals);
}

}

#endif

#endif
